/*
 * Coloring.java
 *
 * Created on May 29, 2006, 3:19 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm;
import java.io.*;
import javax.swing.*;
import java.awt.Color;

/**
 *
 * @author Erik J. Tollerud
 */
public class Coloring {
    
    private enum ColorScheme{UNIFORM, LINEARVAL, LINEARVAL_INV, LINEARSAT, LINEARSAT_INV, PSYCHADELIC, DIRECT_MAP, PACKED_MAP, SIGN_BINARY, INDEX_VALUE};
    
    private ColorScheme cs;
    private float[][] colorMap;
    private float[] directIndecies;
    private float[] colorHolder;
    private float top;
    private float bottom;
    private final float psychOptions[] = {360.0f,270.0f,2.0f};
    private float[] baseColor; // For color schemes (UNIFORM, LINEARVAL, LINEARSAT, SIGN_BINARY) that use a base color to start from
    private boolean zeroBased;
    
    /** Creates a new instance of Coloring on the default Psychadelic Coloring Scheme **/
    public Coloring() {
        this.cs = ColorScheme.PSYCHADELIC;
        this.colorHolder = new float[4];
        this.top = 1.0f;
        this.bottom = 0.0f;
        this.zeroBased = false;
    }
    
    public Coloring(File f, boolean directMap) throws FileNotFoundException, IOException, FileFormatException {
        this();
        this.setColorSchemeToColorMap(f,directMap);
    }
    
    public Coloring(float[][] colorMap, boolean directMap) {
        this();
        this.setColorSchemeToColorMap(colorMap,directMap);
    }
    
    public Coloring(Color uniformColor) {
        this();
        this.setColorSchemeToUniform(uniformColor);
    }
    
    public Coloring(Color positive, Color negative) {
        this();
        this.setColorSchemeToSignBinary(positive,negative);
    }
    
    public void setZeroBased(boolean zeroBased) {
        this.zeroBased = zeroBased;
    }
    
    public boolean isZeroBased() {
        return this.zeroBased;
    }
    
    public void setColorSchemeToIndexValue() {
        this.cs = ColorScheme.INDEX_VALUE;
        this.colorMap = null;
        this.baseColor = null;
        this.directIndecies = null;
    }
    
    public boolean isIndexValue() {
        return (this.cs == ColorScheme.INDEX_VALUE);
    }
    
    public void setColorSchemeToUniform(Color color) {
        this.cs = ColorScheme.UNIFORM;
        baseColor = color.getRGBComponents(null);
        this.colorMap = null;
        this.directIndecies = null;
        
        if(colorHolder != null) //preserves "copy" state
            colorHolder = baseColor;
    }
    
    public boolean isUniform() {
        return (cs == cs.UNIFORM);
    }
    
    public void setColorSchemeToLinearBrightnessGradient(Color color) {
        this.setColorSchemeToLinearValueGradient(color,false);
    }
    
    public void setColorSchemeToLinearBrightnessGradient(Color color, boolean invert) {
        this.setColorSchemeToLinearValueGradient(color,invert);
    }
    
    public boolean isLinearBrightnessGradient() {
        return (cs == cs.LINEARVAL || cs == cs.LINEARVAL_INV);
    }
    
    public void setColorSchemeToLinearValueGradient(Color color) {
        this.setColorSchemeToLinearValueGradient(color,false);
    }
    
    public void setColorSchemeToLinearValueGradient(Color color, boolean invert) {
        if (invert)
            this.cs = ColorScheme.LINEARVAL_INV;
        else
            this.cs = ColorScheme.LINEARVAL;
        this.baseColor = new float[4];
        this.colorMap = null;
        this.directIndecies = null;
        
        color.getRGBComponents(baseColor);
        Coloring.RGBtoHSV(baseColor); //only alters the first 3
        /*//Using Java Color class
        Color.RGBtoHSB(color.getRed(),color.getGreen(),color.getBlue(),baseColor);
        baseColor[3] = color.getAlpha()/255.0f;*/
    }
    
    public boolean isLinearValueGradient() {
        return (cs == cs.LINEARVAL || cs == cs.LINEARVAL_INV);
    }
    
    public void setColorSchemeToLinearSaturationGradient(Color color) {
        this.setColorSchemeToLinearSaturationGradient(color,false);
    }
    public void setColorSchemeToLinearSaturationGradient(Color color, boolean invert) {
        if (invert)
            this.cs = ColorScheme.LINEARSAT_INV;
        else
            this.cs = ColorScheme.LINEARSAT;
        this.baseColor = new float[4];
        this.colorMap = null;
        this.directIndecies = null;
        
        color.getRGBComponents(baseColor);
        Coloring.RGBtoHSV(baseColor); //only alters the first 3
        
        /*//Using Java Color class
        Color.RGBtoHSB(color.getRed(),color.getGreen(),color.getBlue(),baseColor);
        baseColor[3] = color.getAlpha()/255.0f;*/
    }
    
    public boolean isLinearSaturationGradient() {
        return ( cs == cs.LINEARSAT || cs == cs.LINEARSAT_INV);
    }
    
    public void setColorSchemeToPsychadelic() {
        setColorSchemeToPsychadelic(360,270,2);
    }
    
    public boolean isPsychadelic() {
        return ( cs == cs.PSYCHADELIC );
    }
    
    public void setColorSchemeToPsychadelic(float scale, float offset, float satPower) {
        this.cs = ColorScheme.PSYCHADELIC;
        psychOptions[0] = scale;
        psychOptions[1] = offset;
        psychOptions[2] = satPower;
        this.colorMap = null;
        this.baseColor = null;
        this.directIndecies = null;
        
    }
    
    public void setColorSchemeToColorMap() throws SDDMException {
        if(this.colorMap == null)
            throw new SDDMException("No previously defined color map");
        
        if (this.colorMap[0].length == 5)
            this.cs = ColorScheme.DIRECT_MAP;
        else
            this.cs = ColorScheme.PACKED_MAP;
        
        this.baseColor = null;
        
    }
    
    public boolean isColorMap() {
        return (cs == cs.DIRECT_MAP || cs == cs.PACKED_MAP);
    }
    
    public boolean isDirectColorMap() {
        return ( cs == cs.DIRECT_MAP );
    }
    
    /**
     * copies in from supplied color map - packed in ascending order
     * Format: {R,G,B,[A]},{R,G,B,[A]},... or if direct,  {Index,R,G,B,[A]},{Index,R,G,B,[A]},...
     * @param colorMap
     * @param directMap
     * @throws SDDMException if the color map is not the correct format
     */
    public void setColorSchemeToColorMap(float[][] colorMap, boolean directMap) throws SDDMException{
        this.colorMap = new float[colorMap.length][4];
        this.baseColor = null;
        if (directMap)  {
            this.cs = ColorScheme.DIRECT_MAP;
            
            this.directIndecies = new float[colorMap.length];
            for (int i = 0; i < colorMap.length; ++i) {
                switch(colorMap[i].length) {
                    case 5:
                        this.directIndecies[i] = colorMap[i][0];
                        this.colorMap[i][0] = colorMap[i][1];
                        this.colorMap[i][1] = colorMap[i][2];
                        this.colorMap[i][2] = colorMap[i][3];
                        this.colorMap[i][3] = colorMap[i][4];
                    case 4:
                        this.directIndecies[i] = colorMap[i][0];
                        this.colorMap[i][0] = colorMap[i][1];
                        this.colorMap[i][1] = colorMap[i][2];
                        this.colorMap[i][2] = colorMap[i][3];
                        this.colorMap[i][3] = 1.0f;
                        break;
                    default:
                        throw new SDDMException("Direct color map element is not 4 or 5 components");
                }
            }
            sortDirectMap();
        } else {
            this.cs = ColorScheme.PACKED_MAP;
            for (int i = 0; i < colorMap.length; ++i) {
                switch(colorMap[i].length) {
                    case 4:
                        this.colorMap[i][0] = colorMap[i][0];
                        this.colorMap[i][1] = colorMap[i][1];
                        this.colorMap[i][2] = colorMap[i][2];
                        this.colorMap[i][3] = colorMap[i][3];
                    case 3:
                        this.colorMap[i] = new float[5];
                        this.colorMap[i][0] = colorMap[i][0];
                        this.colorMap[i][1] = colorMap[i][1];
                        this.colorMap[i][2] = colorMap[i][2];
                        this.colorMap[i][3] = 1.0f;
                        break;
                    default:
                        throw new SDDMException("Packed color map element is not 3 or 4 components");
                }
            }
        }
    }
    
    /**
     * Must all be RGBA rather than RGB format
     * gets placed in ascending order
     * @param f
     * @param directMap
     * @throws java.io.FileNotFoundException
     * @throws java.io.IOException
     * @throws sddm.FileFormatException
     */
    public void setColorSchemeToColorMap(File f, boolean directMap) throws FileNotFoundException, IOException, FileFormatException {
        //TODO: find a way to specify just 3 components (i.e. RGB instead of RGBA)
        int mapSize = directMap ? 5 : 4;
        this.baseColor = null;
        
        java.util.Scanner sc = new java.util.Scanner(f);
        int count = 0;
        while(sc.hasNextFloat())  {
            sc.nextFloat();
            ++count;
        }
        
        int numOfLines = count/mapSize;
        if (count%mapSize != 0)
            throw new FileFormatException("Number of elements in color map wasn't divisible by "+mapSize);
        
        //close the scanner just in case
        sc.close();
        
        colorMap =  new float[numOfLines][4];
        //include the index if its a direct map
        if (directMap)
            directIndecies = new float[numOfLines];
        
        sc = new java.util.Scanner(f);
        //even though these cases are very similar, one decision point is better (maybe java compiler does this anyway?)
        if (directMap) {
            this.cs = ColorScheme.DIRECT_MAP;
            for (int i = 0; i < numOfLines; ++i) {
                directIndecies[i] = sc.nextFloat();
                colorMap[i][0] = sc.nextFloat();
                colorMap[i][1] = sc.nextFloat();
                colorMap[i][2] = sc.nextFloat();
                colorMap[i][3] = sc.nextFloat();
            }
            sortDirectMap();
        } else {
            this.cs = ColorScheme.PACKED_MAP;
            for (int i = 0; i < numOfLines; ++i) {
                colorMap[i][0] = sc.nextFloat();
                colorMap[i][1] = sc.nextFloat();
                colorMap[i][2] = sc.nextFloat();
                colorMap[i][3] = sc.nextFloat();
            }
        }
        sc.close();
    }
    
    public boolean isSignBinary() {
        return ( cs == cs.SIGN_BINARY );
    }
    
    public void setColorSchemeToSignBinary(Color positive, Color negative) {
        //using directIndecies to hold negative colors, baseColor for positive
        this.cs = ColorScheme.SIGN_BINARY;
        this.directIndecies = new float[4];
        this.baseColor = new float[4];
        this.colorMap = null;
        
        positive.getComponents(this.baseColor);
        negative.getComponents(this.directIndecies);
    }
    
    public ColorScheme getColorSchemeType() {
        return this.cs;
    }
    
    /**
     * Determines the color using a method determined by the color map type.  Same as getColorByRangedValue for
     * psychadelic and non-direct maps, or getColorByDirectValue for direct maps
     * @param value the index to use to determine a color
     * @return an {R,G,B,A} array. If copyColor is true, it is a new array.
     */
    public float[] getColorByValue(float value ) {
        float[] result = null;
        switch(this.cs) {
            case UNIFORM:
            case LINEARSAT:
            case LINEARSAT_INV:
            case LINEARVAL:
            case LINEARVAL_INV:
            case PSYCHADELIC:
            case PACKED_MAP:
                result = this.getColorByRangedValue(value);
                break;
            case DIRECT_MAP:
            case SIGN_BINARY:
            case INDEX_VALUE:
                result = this.getColorByDirectValue(value);
                break;
                
        }
        return result;
    }
    
    /**
     * This determines the color based on this Coloring object's settings and a value. The value is expected to be scaled into [0,1] before being passed in. <br>
     * For maps, this will use the closest index, or the top or bottom color if the value is above or below the index.
     * @param value a value on [0,1].
     * @return a 4-element array in {R,G,B,A} format. May not be stable over state changes in this Coloring object if copyColor is false
     */
    public float[] getColorByScaledValue(float value)  {
        float[] result = null; //stubborn java compiler refuses to admit that I took care of all cases, so this initializes to null
        boolean copy = (colorHolder == null);
        boolean invert = false;
        
        switch (cs) {
            case UNIFORM:
                if (copy) {
                    result = new float[4];
                    result[0] = baseColor[0];
                    result[1] = baseColor[1];
                    result[2] = baseColor[2];
                    result[3] = baseColor[3];
                } else
                    result = colorHolder;
                
                break;
            case LINEARSAT_INV:
                invert = true;
            case LINEARSAT:
                if (copy)
                    result = new float[4];
                else
                    result = colorHolder;
                
                result[0] = baseColor[0];
                result[1] = invert ? 1.0f - value : value;
                result[2] = baseColor[2];
                result[3] = baseColor[3];
                Coloring.HSVtoRGB(result);
                /*//Uses Java Color class
                int rgb = Color.HSBtoRGB(baseColor[0],value,baseColor[2]);
                result[0] = (rgb >> 16)& 0xff;
                result[1] = (rgb >> 8) & 0xff;
                result[2] = rgb & 0xff;
                result[3] = baseColor[3];*/
                break;
            case LINEARVAL_INV:
                invert = true;
            case LINEARVAL:
                if (copy)
                    result = new float[4];
                else
                    result = colorHolder;
                
                result[0] = baseColor[0];
                result[1] = baseColor[1];
                result[2] = invert ? 1.0f - value : value;
                result[3] = baseColor[3];
                Coloring.HSVtoRGB(result);
                /*//Uses Java Color class
                rgb = Color.HSBtoRGB(baseColor[0],baseColor[1],value);
                result[0] = (rgb >> 16)& 0xff;
                result[1] = (rgb >> 8) & 0xff;
                result[2] = rgb & 0xff;
                result[3] = baseColor[3];*/
                break;
            case PSYCHADELIC:
                if (copy) {
                    result = new float[4];
                    result[3] = 1.0f;
                } else
                    result = colorHolder;
                psychadelicColorFunc(value, result,psychOptions[0],psychOptions[1],psychOptions[2]);
                
                break;
            case DIRECT_MAP:
                //use a value scaled based on the range of the indecies for the color map; 0->the bottom index, and 1->the top
                float scaledValue = value*(directIndecies[colorMap.length-1]-directIndecies[0])+directIndecies[0];
                
                int indx;
                if (scaledValue < directIndecies[0])
                    indx = 0;
                else if (scaledValue > directIndecies[directIndecies.length-1])
                    indx = colorMap.length-1;
                else {
                    //map is in ascending order, so we can just search up through the values
                    //TODO: use binary search if this turns out to be a CPU hog
                    indx = 0; //can skip 0 because if its less than the first element, its already been taken care of above
                    while(directIndecies[++indx] < scaledValue);
                    
                    //pick either that one or the one below, depending on which is closest
                    if (directIndecies[indx] - scaledValue > scaledValue - directIndecies[indx-1])
                        --indx;
                    
                    result = colorMap[indx];
                }
                
                if (copy) {
                    result = new float[4];
                    result[0] = colorMap[indx][0];
                    result[1] = colorMap[indx][1];
                    result[2] = colorMap[indx][2];
                    result[3] = colorMap[indx][3];
                } else
                    result = colorMap[indx];
                break;
            case PACKED_MAP:
                indx = (int) (value*(colorMap.length - 1));
                if (copy) {
                    result = new float[4];
                    result[0] = colorMap[indx][0];
                    result[1] = colorMap[indx][1];
                    result[2] = colorMap[indx][2];
                    result[3] = colorMap[indx][3];
                } else
                    result = colorMap[indx];
                break;
            case SIGN_BINARY:
                //baseColor has positive color
                //directIndicies has negative color
                if (copy) {
                    result = new float[4];
                    if(value < 0.5) {
                        result[0] = this.directIndecies[0];
                        result[1] = this.directIndecies[1];
                        result[2] = this.directIndecies[2];
                        result[3] = this.directIndecies[3];
                    } else {
                        result[0] = this.baseColor[0];
                        result[1] = this.baseColor[1];
                        result[2] = this.baseColor[2];
                        result[3] = this.baseColor[3];
                    }
                    
                } else {
                    if(value < 0.5)
                        result = this.directIndecies;
                    else
                        result = this.baseColor;
                }
                break;
            case INDEX_VALUE:
                result = this.getColorByDirectValue(value);
                break;
        }
        return result;
    }
    
    /**
     * Determines the color by direct indexing (or the closest index, if match is not exact) if this is a direct map.
     * If the value is above or below the indexed range, the color will be the upper or lower index color.<br>
     * If this is not a direct map, the value will be applied as a scaled value over the whole range of integers.
     * @param value an index for the color map
     * @return an {R,G,B,A} array. If copyColor is true, it is a new array.
     */
    public float[] getColorByDirectValue(int value){
        float[] result = null;
        switch(cs) {
            case UNIFORM:
            case LINEARSAT:
            case LINEARSAT_INV:
            case LINEARVAL:
            case LINEARVAL_INV:
            case PACKED_MAP:
            case PSYCHADELIC:
                result = getColorByScaledValue(((value-Integer.MIN_VALUE)/Integer.MAX_VALUE)/2.0f);
                break;
            case DIRECT_MAP:
            case SIGN_BINARY:
                //TODO: test if there are problems with this conversion
                result = getColorByDirectValue((float)value);
                break;
            case INDEX_VALUE:
                if (colorHolder == null)
                    result = new float[4];
                else
                    result = colorHolder;
                result[0]= ((value >> 24) & 0xff)/255.0f;
                result[1]= ((value >> 16) & 0xff)/255.0f;
                result[2]= ((value >> 8) & 0xff)/255.0f;
                result[3]= (value & 0xff)/255.0f;
                break;
        }
        return result;
    }
    
    /**
     * Determines the color by direct indexing (or the closest index, if match is not exact) if this is a direct map.
     * If the value is above or below the indexed range, the color will be the upper or lower index color.<br>
     * If this is not a direct map, the value will be applied as a scaled value over the whole range of floats.
     * @param value an index for the color map
     * @return an {R,G,B,A} array. If copyColor is true, it is a new array.
     */
    public float[] getColorByDirectValue(float value){
        
        float[] result = null;
        
        switch(cs) {
            case UNIFORM:
            case LINEARSAT:
            case LINEARSAT_INV:
            case LINEARVAL:
            case LINEARVAL_INV:
            case PACKED_MAP:
            case PSYCHADELIC:
                result = getColorByScaledValue(((value-Float.MIN_VALUE)/Float.MAX_VALUE)/2.0f);
                break;
            case DIRECT_MAP:
                int indx;
                if (value < directIndecies[0])
                    indx = 0;
                else if (value > directIndecies[directIndecies.length-1])
                    indx = colorMap.length-1;
                else {
                    //map is in ascending order, so we can just search up through the values
                    //TODO: use binary search if this turns out to be a CPU hog
                    indx = 0; //can skip 0 because if its less than the first element, its already been taken care of above
                    while(directIndecies[++indx] < value);
                    
                    //match the closest color, if required
                    if (directIndecies[indx] != value) {
                        // if the element one back down is closer in value, use it instead
                        if (directIndecies[indx] - value > value - directIndecies[indx-1])
                            --indx;
                    }
                }
                
                //copy or link, depending on whether or not the holder is instantiated (used for psychadelic or as a marker
                if (colorHolder == null) {
                    result = new float[4];
                    result[0] = colorMap[indx][0];
                    result[1] = colorMap[indx][1];
                    result[2] = colorMap[indx][2];
                    result[3] = colorMap[indx][3];
                } else
                    result = colorMap[indx];
                
                break;
            case SIGN_BINARY:
                //baseColor has positive color
                //directIndicies has negative color
                if (colorHolder == null) {
                    result = new float[4];
                    if(value < 0) {
                        result[0] = this.directIndecies[0];
                        result[1] = this.directIndecies[1];
                        result[2] = this.directIndecies[2];
                        result[3] = this.directIndecies[3];
                    } else {
                        result[0] = this.baseColor[0];
                        result[1] = this.baseColor[1];
                        result[2] = this.baseColor[2];
                        result[3] = this.baseColor[3];
                    }
                    
                } else {
                    if(value < 0)
                        result = this.directIndecies;
                    else
                        result = this.baseColor;
                }
                break;
            case INDEX_VALUE:
                result = getColorByDirectValue(Float.floatToIntBits(value));
                break;
        }
        return result;
    }
    
    /**
     * Determines the color using a linear mapping between the values specified in the setTop and setBottom methods.
     * @param value the value to use in determining the color, should be between this object's top and bottom (although range checking is not performed for performance).
     * @return an {R,G,B,A} array.  If copyColor is true, it is a new array.
     */
    public float[] getColorByRangedValue(float value) {
        if(this.zeroBased) {
            if (value < 0)
                return  getColorByScaledValue( value / bottom);
            else
                return  getColorByScaledValue( value / top );
        } else
            //fits the result into [0,1] based on top and bottom
            return getColorByScaledValue((value-bottom) / (top-bottom));
    }
    
    public float getTop(){
        return top;
    }
    
    public void setTop(float top) {
        if(top == this.bottom)
            throw new SDDMException("Tried to set top to same value as bottom");
        this.top = top;
    }
    
    public float getBottom() {
        return bottom;
    }
    
    public void setBottom(float bottom) {
        if(this.top == bottom)
            throw new SDDMException("Tried to set borrom to same value as top");
        this.bottom = bottom;
    }
    
    public void setCopyResult(boolean shouldCopy) {
        
        if (shouldCopy)
            this.colorHolder = null;
        else
            this.colorHolder = new float[4];
        if (cs == cs.UNIFORM)
            this.colorHolder = this.baseColor;
        
    }
    
    /**
     * Determines if the component values are copied or placed in a temporary storage array each time a color is computed.
     * @return true if color values are copied(always false with the uniform color scheme)
     */
    public boolean isCopyResult() {
        return (colorHolder == null);
    }
    
    
    
    /**
     * Generates a set of swing dialogs to configure a new Coloring object.
     *
     * @return the Coloring object configured by the dialogs, or null if something is cancelled
     * @param parent the component to use as the parent for the JOptionPane dialogs
     * @throws java.awt.HeadlessException if GraphicsEnvironment.isHeadless returns true
     */
    public static Coloring coloringOptionPane(java.awt.Component parent) throws java.awt.HeadlessException {
        return coloringOptionPane(parent,null);
    }
    
    /**
     * Generates a set of swing dialogs to configure a new Coloring object based on an earlier Coloring object.
     *
     *
     * @return the Coloring object configured by the dialogs, or null if something is cancelled
     * @param initial the initial state to use in the dialogs
     * @param parent the component to use as the parent for the JOptionPane dialogs
     * @throws java.awt.HeadlessException if GraphicsEnvironment.isHeadless returns true
     */
    public static Coloring coloringOptionPane(java.awt.Component parent, Coloring initial) throws java.awt.HeadlessException {
        Coloring c = null;
        
        JRadioButton uniformButton = new JRadioButton("Uniform");
        JRadioButton linearSatButton = new JRadioButton("Linear Map by Saturation");
        JRadioButton linearValButton = new JRadioButton("Linear Map by Value/Brightness");
        JRadioButton psychButton = new JRadioButton("Psychadelic");
        JRadioButton colorMapButton = new JRadioButton("Color Map");
        JRadioButton directButton = new JRadioButton("Direct Color Map");
        JRadioButton binaryButton = new JRadioButton("Sign Binary");
        JCheckBox zeroBaseCheckBox = new JCheckBox("Zero-Base Ranging");
        
        ButtonGroup grp = new ButtonGroup();
        grp.add(uniformButton);
        grp.add(linearSatButton);
        grp.add(linearValButton);
        grp.add(psychButton);
        grp.add(colorMapButton);
        grp.add(directButton);
        grp.add(binaryButton);
        
        JPanel panel = new JPanel();
        panel.add(uniformButton);
        panel.add(linearSatButton);
        panel.add(linearValButton);
        panel.add(psychButton);
        panel.add(colorMapButton);
        panel.add(directButton);
        panel.add(binaryButton);
        panel.add(zeroBaseCheckBox);
        
        if (initial != null) {
            if (initial.isLinearSaturationGradient())
                linearSatButton.setSelected(true);
            else if (initial.isLinearValueGradient())
                linearValButton.setSelected(true);
            else if(initial.isPsychadelic())
                psychButton.setSelected(true);
            else if (initial.isColorMap())
                colorMapButton.setSelected(true);
            else if (initial.isDirectColorMap())
                directButton.setSelected(true);
            else if (initial.isSignBinary())
                binaryButton.setSelected(true);
            else
                uniformButton.setSelected(true);
        } else
            uniformButton.setSelected(true);
        
        net.java.joglutils.JPanelDialog jpd = new net.java.joglutils.JPanelDialog(panel);
        jpd.setTitle("Coloring Schemes");
        
        if(jpd.showAsModal()) {
            c = new Coloring();
            if (uniformButton.isSelected()||linearSatButton.isSelected()||linearValButton.isSelected()) {
                Color beforeColor = null;
                if (initial != null)
                    beforeColor = (initial.isUniform() || initial.isLinearValueGradient()|| initial.isLinearSaturationGradient()) ?
                        new Color(initial.baseColor[0],initial.baseColor[1],initial.baseColor[2],initial.baseColor[3])
                        : null;
                
                String title;
                if (linearSatButton.isSelected())
                    title = "Linear Saturation Gradient Base Color";
                else if (linearValButton.isSelected())
                    title = "Linear Value/Brightness Gradient Base Color";
                else
                    title = "Unifrom Color";
                
                Color result = JColorChooser.showDialog(parent,title,beforeColor);
                if (result != null) {
                    if (linearValButton.isSelected()) {
                        boolean b = (JOptionPane.YES_OPTION==JOptionPane.showConfirmDialog(parent,"Invert value/brightness gradient (i.e. Color->Black)","Invert?",JOptionPane.YES_NO_OPTION));
                        c.setColorSchemeToLinearValueGradient(result,b);
                    } else if (linearSatButton.isSelected()) {
                        boolean b = (JOptionPane.YES_OPTION==JOptionPane.showConfirmDialog(parent,"Invert saturation gradient (i.e. Color->White)","Invert?",JOptionPane.YES_NO_OPTION));
                        c.setColorSchemeToLinearSaturationGradient(result,b);
                    } else
                        c.setColorSchemeToUniform(result);
                }
                
                else
                    c = null; //do nothing if cancelled
            } else if (psychButton.isSelected()) {
                try {
                    float scale = Float.parseFloat(JOptionPane.showInputDialog(parent,"Psychadelic hue scaling?","360"));
                    float offset = Float.parseFloat(JOptionPane.showInputDialog(parent,"Psychadelic hue offset?","270"));
                    float power = Float.parseFloat(JOptionPane.showInputDialog(parent,"Psychadelic value exponent power?","2"));
                    c.setColorSchemeToPsychadelic(scale,offset,power);
                } catch (NullPointerException e) {
                    //If any dialogs are cancelled, do nothing
                    c = null;
                }
                
                
            } else if (binaryButton.isSelected()) {
                Color initialPos = new Color(1.0f,0,0,1.0f);
                Color initialNeg = new Color(0,0,1.0f,1.0f);
                if (initial != null) {
                    initialPos = new Color(initial.baseColor[0],initial.baseColor[1],initial.baseColor[2],initial.baseColor[3]);
                    initialNeg = new Color(initial.directIndecies[0],initial.directIndecies[1],initial.directIndecies[2],initial.directIndecies[3]);
                }
                
                Color pos = JColorChooser.showDialog(parent,"Color for Positive Values?",initialPos);
                Color neg = null;
                if (pos != null)
                    neg = JColorChooser.showDialog(parent,"Color for Negative Values?",initialNeg);
                
                if (neg != null)
                    c = new Coloring(pos,neg);
                else
                    c = null;
            } else  {
                //Color map file buttons
                boolean direct = directButton.isSelected();
                
                JFileChooser jfc = new JFileChooser();
                jfc.setMultiSelectionEnabled(false);
                if (jfc.showOpenDialog(parent) == jfc.APPROVE_OPTION) {
                    try {
                        c = new Coloring(jfc.getSelectedFile(),direct);
                    } catch(Exception e){
                        JOptionPane.showMessageDialog(parent,"Problem loading color map file:"+e.getMessage(),"Error loading color Map",JOptionPane.ERROR_MESSAGE);
                    }
                } else
                    c = null;
            }
            if (c != null)
                c.setZeroBased(zeroBaseCheckBox.isSelected());
        }
        //else do nothing - leave c null
        return c;
    }
    
    public static Coloring makeColoringFromArgs(java.util.HashMap<String,Integer> switchMap, String[] args) throws ParsingException{
        //[-uniform] r g b
        //[-satgradient] r g b [(true|t|invert)|(false|f)]
        //[-valgradient|brightgradient] r g b [(true|t|invert)|(false|f)]
        //[-psychadelic] [scale offset power]
        //[-colormap|map] filename [(true|t|direct)|(false|f)]    (defaults to indirect/false)
        
        Coloring c = null;
        
        Integer uIndx = switchMap.remove("-uniform");
        if (uIndx != null) {
            try{
                float r = Float.parseFloat(args[uIndx+1]);
                float g = Float.parseFloat(args[uIndx+2]);
                float b = Float.parseFloat(args[uIndx+3]);
                c = new Coloring(new Color(r,g,b));
            } catch (Exception e) {
                throw new ParsingException("Problem making uniform coloring from command line:"+e.getMessage(),e);
            }
        }
        
        Integer satGradIndx = switchMap.remove("-satgradient");
        if (satGradIndx != null) {
            try{
                float r = Float.parseFloat(args[satGradIndx+1]);
                float g = Float.parseFloat(args[satGradIndx+2]);
                float b = Float.parseFloat(args[satGradIndx+3]);
                
                boolean invert = false;
                if (satGradIndx+4 == args.length || switchMap.containsValue(satGradIndx+4));//leave it as the default
                else {
                    String invertStr = args[satGradIndx+4];
                    if(invertStr.matches("true|t|invert"))
                        invert = true;
                    else if (invertStr.matches("false|f"))
                        invert = false;
                    else
                        throw new ParsingException("Invalid option for color gradient:"+invertStr);
                }
                
                c = new Coloring();
                c.setColorSchemeToLinearSaturationGradient(new Color(r,g,b),invert);
            } catch (Exception e) {
                throw new ParsingException("Problem making uniform coloring from command line:"+e.getMessage(),e);
            }
        }
        
        Integer valGradIndx = switchMap.remove("-valgradient");
        if (valGradIndx == null)
            valGradIndx = switchMap.remove("-brightgradient");
        if (valGradIndx != null) {
            try{
                float r = Float.parseFloat(args[valGradIndx+1]);
                float g = Float.parseFloat(args[valGradIndx+2]);
                float b = Float.parseFloat(args[valGradIndx+3]);
                
                boolean invert = false;
                if (valGradIndx+4 == args.length || switchMap.containsValue(valGradIndx+4));//leave it as the default
                else {
                    String invertStr = args[valGradIndx+4];
                    if(invertStr.matches("true|t|invert"))
                        invert = true;
                    else if (invertStr.matches("false|f"))
                        invert = false;
                    else
                        throw new ParsingException("Invalid option for color gradient:"+invertStr);
                }
                
                c = new Coloring();
                c.setColorSchemeToLinearSaturationGradient(new Color(r,g,b),invert);
            } catch (Exception e) {
                throw new ParsingException("Problem making uniform coloring from command line:"+e.getMessage(),e);
            }
        }
        
        
        Integer mapIndx = switchMap.remove("-colormap");
        if (mapIndx == null)
            mapIndx = switchMap.remove("-map");
        if (mapIndx != null) {
            
            String filename = args[mapIndx+1];
            java.io.File mapFile = new java.io.File(filename);
            
            boolean direct = false;
            if (mapIndx+2 == args.length || switchMap.containsValue(mapIndx+2));//leave it as the default
            else {
                String directStr = args[mapIndx+2];
                if(directStr.matches("true|t|direct"))
                    direct = true;
                else if (directStr.matches("false|f"))
                    direct = false;
                else
                    throw new ParsingException("Invalid option for color map:"+directStr);
            }
            try{
                c = new Coloring(mapFile,direct);
            } catch (Exception e) {
                throw new ParsingException("Exception while loading color map: "+e.getMessage(),e);
            }
        }
        
        Integer psychIndx = switchMap.remove("-psychadelic");
        if (psychIndx != null) {
            c = new Coloring();
            try {
                float scale = Float.parseFloat(args[psychIndx+1]);
                float offset = Float.parseFloat(args[psychIndx+2]);
                float power = Float.parseFloat(args[psychIndx+3]);
                c.setColorSchemeToPsychadelic(scale,offset,power);
            } catch(Exception e) {
                c.setColorSchemeToPsychadelic();
            }
        }
        return c;
    }
    
    private void sortDirectMap() {
        //put all elements a sorted map
        java.util.TreeMap<Float,float[]> map = new java.util.TreeMap<Float,float[]>();
        for(float[] f : this.colorMap)
            map.put(f[0],f);
        
        //Return the components to the array in ascending order
        int i = 0;
        for(float[] f : map.values())
            this.colorMap[i++] = f;
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(float) in index mode)
     * @param rgbaComp {r,g,b,a} with each component on [0,1]
     * @return float value provided as index value provided from a Coloring object in INDEX_VALUE mode
     */
    public static float getIndexedFloat(float[] rgbaComp) {
        return Float.intBitsToFloat(getIndexedInt(rgbaComp));
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(float) in index mode)
     * @param r on [0,1]
     * @param g on [0,1]
     * @param b on [0,1]
     * @param a on [0,1]
     * @return float value provided as index value provided from a Coloring object in INDEX_VALUE mode
     */
    public static float getIndexedFloat(float r, float g, float b, float a) {
        final float[] components = {r,g,b,a};
        return getIndexedFloat(components);
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(float) in index mode)
     * @param r on [0,255]
     * @param g on [0,255]
     * @param b on [0,255]
     * @param a on [0,255]
     * @return float value provided as index value provided from a Coloring object in INDEX_VALUE mode
     */
    public static float getIndexedFloat(short r, short g, short b, short a) {
        final float[] components = {r/255.0f,g/255.0f,b/255.0f,a/255.0f};
        return getIndexedFloat(components);
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(int) in index mode)
     * @param rgbaComp {r,g,b,a} with each component on [0,1]
     * @return the 32-bit int indexed by the value provided from a Coloring object in INDEX_VALUE mode
     */
    public static int getIndexedInt(float[] rgbaComp) {
        final int r = (int)(rgbaComp[0]*255.0f);
        final int g = (int)(rgbaComp[1]*255.0f);
        final int b = (int)(rgbaComp[2]*255.0f);
        final int a = (int)(rgbaComp[3]*255.0f);
        return (r << 24) + (g << 16) + (b << 8) + a;
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(int) in index mode)
     * @param r on [0,1]
     * @param g on [0,1]
     * @param b on [0,1]
     * @param a on [0,1]
     * @return the 32-bit int indexed by the value provided from a Coloring object in INDEX_VALUE mode
     */
    public static int getIndexedInt(float r, float g, float b, float a) {
        final float[] components = {r,g,b,a};
        return getIndexedInt(components);
    }
    
    /**
     * Retrieves the value used to index the provided color (paired with getColorByDirectValue(int) in index mode)
     * @param r on [0,255]
     * @param g on [0,255]
     * @param b on [0,255]
     * @param a on [0,255]
     * @return the 32-bit int indexed by the value provided from a Coloring object in INDEX_VALUE mode
     */
    public static int getIndexedInt(short r, short g, short b, short a) {
        final float[] components = {r/255.0f,g/255.0f,b/255.0f,a/255.0f};
        return getIndexedInt(components);
    }
    
    //----------------static utility methods below-----------
    //used in timing test
    //static long start = 0;
    //static long end;
    //static int staticCount = 0;
    public static void HSVtoRGB(float[] colorArray) {
        
        //Timing test
        /*if(staticCount == 0)
            start = System.nanoTime();
        staticCount++;
        if (staticCount%100000 == 0) {
            end = System.nanoTime();
            System.err.println("count:"+staticCount);
            System.err.println("Time:"+((end-start)/(10E9))+" s");
            staticCount = 0;
        }*/
        //The Color method is a bit less efficient than the customized implementation
        /*int color = java.awt.Color.HSBtoRGB(colorArray[0],colorArray[1],colorArray[2]);
        colorArray[2] = color & 0xff;
        colorArray[1] = (color >> 8) & 0xff;
        colorArray[0] =  (color  >> 16)  & 0xff; */
        //general formula stolen from wikipedia entry on HSV color space
        int Hi;
        float f,p,q,t,R=0,G=0,B=0;
        float H = colorArray[0];
        float S = colorArray[1];
        float V = colorArray[2];
        
        if (S==0.0f) {
            R=V;
            G=V;
            B=V;
        } else {
            while(H>360)
                H=H-360;
            while(H<0)
                H=H+360;
            Hi=(int)(Math.floor(H/60)) % 6;
            f=(H/60)-Hi;
            p=V*(1-S);
            q=V*(1-f*S);
            t=V*(1-(1-f)*S);
            switch (Hi) {
                case 0:
                    R=V;
                    G=t;
                    B=p;
                    break;
                case 1:
                    R=q;
                    G=V;
                    B=p;
                    break;
                case 2:
                    R=p;
                    G=V;
                    B=t;
                    break;
                case 3:
                    R=p;
                    G=q;
                    B=V;
                    break;
                case 4:
                    R=t;
                    G=p;
                    B=V;
                    break;
                case 5:
                    R=V;
                    G=p;
                    B=q;
                    break;
            }
        }
        colorArray[0]=R;
        colorArray[1]=G;
        colorArray[2]=B;
    }
    
    public static void RGBtoHSV( float[] colorArray ) {
        
        // RGB are each on [0, 1]. S and V are returned on [0, 1] and H is
        // returned on [0, 6]. Exception: H is returned UNDEFINED if S==0.
        float R = colorArray[0], G = colorArray[1], B = colorArray[2], v, x, f;
        int i;
        
        x = Math.min(R, Math.min(G, B));
        v = Math.max(R, Math.max(G, B));
        
        if(v == x) {
            colorArray[0]=0;
            colorArray[1]=0;
            colorArray[2]=v;
        } else {
            f = (R == x) ? G - B : ((G == x) ? B - R : R - G);
            i = (R == x) ? 3 : ((G == x) ? 5 : 1);
            colorArray[0]=(i - f /(v - x))*60.0f;
            colorArray[1]=(v - x)/v;
            colorArray[2]=v;
        }
        
    }
    
    public static void psychadelicColorFunc(float value, float[] colorArray) {
        psychadelicColorFunc(value,colorArray,360,270,2);
    }
    
    public static void psychadelicColorFunc(float value, float[] colorArray,float scale, float offset) {
        psychadelicColorFunc(value,colorArray,scale,offset,2);
    }
    
    public static void psychadelicColorFunc(float value, float[] colorArray,float scale, float offset, float satPower) {
        colorArray[0]=offset + scale*(value);
        colorArray[1]=2*(float)Math.exp(-2*(float)Math.pow(value,satPower))-0.3f;
        colorArray[2]=1;
        HSVtoRGB(colorArray);
    }
    
}
