/*
 * PointSet.java
 *
 * Created on August 2, 2006, 3:30 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import sddm.*;
import java.io.*;
import java.util.*;

/**
 * 
 * A Set of points
 * @author Erik J. Tollerud
 */
public class PointSet implements Iterable<float[]>, Cloneable{
    //internally, the arrays are all 4-element.
    /**
     * The backing hashSet of this PointSet.
     */
    protected HashSet<float[]> set;
    //format is {xmin,xmax,xavg,ymin,ymax,yavg,zmin,zmax,zavg,wmin,wmax,wavg}
    float[] stats;
    
    
    /** Creates a new instance of PointSet */
    public PointSet() {
        set = new HashSet<float[]>();
    }
    
    /**
     * Creates a new instance of PointSet from the supplied points (eliminating duplicates)
     * @param points an array of 3 or 4-element arrays to use as the points.
     * @throws sddm.SDDMException if any of the points are not 3 or 4 elements
     */
    public PointSet(float[][] points)  throws SDDMException {
        set = new HashSet<float[]>();
        this.setPoints(points);
    }
    
    /**
     * 
     * Creates a new instance of PointSet from a file (see {@link loadPointsFromFile(File)})
     */
    public PointSet(File pointFile) throws FileFormatException, FileNotFoundException {
        set = new HashSet<float[]>();
        this.loadPointsFromFile(pointFile);
    }
    
    /**
     * 
     * Creates a new instance of PointSet from a file (see {@link loadPointsFromFile(File, boolean)})
     */
    public PointSet(File pointFile, boolean threeElement) throws FileFormatException, FileNotFoundException {
        set = new HashSet<float[]>();
        this.loadPointsFromFile(pointFile, threeElement);
    }
    
    PointSet(PointList pl) {
        this();
        float[][] array = new float [1][1];
        array = pl.list.toArray(array);
        this.setPoints(array);
    }
    
    /**
     * Shallowly copies this PointSet to a PointList in iterator order.
     * @return a PointList with the same points as this PointSet.
     */
    public PointList toList() {
        return new PointList(this);
    }
    
    /**
     * Sets the points to the specified set of points (eliminating duplicates).  This clears the existing points.
     * @param points An array of 3 or 4-element arrays specifying the points to use.
     * @throws sddm.SDDMException If any of the points are not 3 or 4 elements long
     */
    public synchronized void setPoints(float[][] points) throws SDDMException {
        set.clear();
        for (float[] f : points)
            this.addPoint(f);
    }
    
    /**
     * 
     * Loads a set of points from a file that has only numbers in x0 y0 z0 x1 y1 z1 x2 y2 z2 ... format.
     * @param pointFile the file from which to load the points.
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3 or 4.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
    public synchronized void loadPointsFromFile(File pointFile) throws FileFormatException, FileNotFoundException {
        loadPointsFromFile(pointFile,true);
    }
    
    /**
     * 
     * Loads a set of points from a file that has only numbers in x0 y0 z0 x1 y1 z1 x2 y2 z2 ...  or x0 y0 z0 w0 x1 y1 z1 w1 x2 y2 z2 w2 ... format.
     * @param pointFile the file from which to load the points.
     * @param threeElement if true, only sets x,y, and z cooordinates from the file (w is set to 1.0).
     * @throws sddm.FileFormatException If the file does not have numbers, or the number of numbers isn't divisible by 3 or 4.
     * @throws java.io.FileNotFoundException If the specified file doesn't exist.
     */
    public synchronized void loadPointsFromFile(File pointFile, boolean threeElement) throws FileFormatException, FileNotFoundException {
        //TODO:implement
        Scanner s = new Scanner(pointFile);
        if (threeElement) {
            while (s.hasNextFloat()) {
                float x = s.nextFloat();
                float y = s.nextFloat();
                //if (s.hasNextFloat())
                //    throw new FileFormatException("File loaded to point set does not have a number of elements divisible by 3.");
                float z = s.nextFloat();
                
                this.addPoint(x,y,z,1.0f);
            }
        } else {
            while (s.hasNextFloat()) {
                float x = s.nextFloat();
                float y = s.nextFloat();
                float z = s.nextFloat();
                //if (s.hasNextFloat())
                //    throw new FileFormatException("File loaded to point set does not have a number of elements divisible by 4.");
                float w = s.nextFloat();
                
                this.addPoint(x,y,z,w);
            }
        }
        stats = null;
    }
    
    /**
     * Adds a point with the specified coordinates to the set
     * @param point An {x,y,z} or {x,y,z,w} array to be copied into the set.
     * @throws sddm.SDDMException if the supplied point does not have 3 or 4 elements
     */
    public synchronized void addPoint(float[] point) throws SDDMException {
        if (point.length == 3)
            this.addPoint(point[0],point[1],point[2]);
        else if (point.length == 4)
            this.addPoint(point[0],point[1],point[2],point[3]);
        else
            throw new SDDMException("Tried to add point that doesn't have 3 or 4 components");
        
        stats = null;
    }
    
    /**
     * 
     * Adds points with the specified coordinates to the set
     * @param points An array of 3 or 4-element arrays to be copied elementwise into this set.
     * @throws sddm.SDDMException If any of the supplied points are not 3 or 4 elements long.
     */
    public synchronized void addPoints(float[][] points) throws SDDMException {
        for (float[] f : points)
            this.addPoint(f);
    }
    
    /**
     * Adds a point with the specified coordinates (and w = 1.0f).  Does nothing if the point already exists in the set.
     * @param x the x-coordinate of the point to be added.
     * @param y the y-coordinate of the point to be added.
     * @param z the z-coordinate of the point to be added.
     */
    public synchronized void addPoint(float x, float y, float z) {
        float[] element= new float[4];
        element[0] = x;
        element[1] = y;
        element[2] = z;
        element[3] = 1.0f;
        
        stats = null;        
        set.add(element);
    }
    
    /**
     * Adds a point with the specified coordinates.  Does nothing if the point already exists in the set.
     * @param x the x-coordinate of the point to be added.
     * @param y the y-coordinate of the point to be added.
     * @param z the z-coordinate of the point to be added.
     * @param w the w-coordinate of the point to be added.
     */
    public synchronized void addPoint(float x, float y, float z, float w) {
        float[] element= new float[4];
        element[0] = x;
        element[1] = y;
        element[2] = z;
        element[3] = w;
        
        stats = null;
        set.add(element);
    }
    
    /**
     * 
     * Determines if the object referenced by the parameter exists in this PointSet
     * @param point a reference to an array that may be in this PointSet
     * @return true if the array is present in this PointSet.
     */
    public boolean contains(float[] point) {
        return set.contains(point);
    }
    
    /**
     * Determines if the point with the coordinates given by the supplied array exists in this PointSet.
     * 
     * @return true if a point in this PointSet has the sepcified coordinates.
     * @param point a 3 or 4 element array to compare elementwise to the points in the set.
     * @throws sddm.SDDMException if the supplied point is not 3 or 4 elements long.
     */
    public boolean containsPoint(float[] point)  throws SDDMException {
        boolean exists = false;
        
        if (point.length == 3) {
            for (float[] f: this.set ) {
                if (f[0] == point[0] &&f[1] == point[1]&&f[2] == point[2])
                    exists = true;
            }
        }
        else if (point.length == 4) {
            for (float[] f: this.set ) {
                if (f[0] == point[0] &&f[1] == point[1]&&f[2] == point[2]&&f[3] == point[3])
                    exists = true;
            }
        }
        else
            throw new SDDMException("Supplied point to test PointSet for doesn't have 3 or 4 elements.");
            
            return exists;
    }
    
    /**
     * 
     * Determines if the point with the specified coordinates exists in this PointSet.
     * @param x the x-coordinate of the point to search for.
     * @param y the y-coordinate of the point to search for.
     * @param z the z-coordinate of the point to search for.
     * @return true if an element exists with coordinates matching the supplied values.
     */
    public boolean containsPoint(float x, float y, float z) {
        boolean exists = false;
        for (float[] f: this.set ) {
                if (f[0] == x && f[1] == y && f[2] == z)
                    exists = true;
            }
        return exists;
    }
    
    /**
     * 
     * Determines if the point with the specified coordinates exists in this PointSet.
     * @param x the x-coordinate of the point to search for.
     * @param y the y-coordinate of the point to search for.
     * @param z the z-coordinate of the point to search for.
     * @param w the w-coordinate of the point to search for.
     * @return true if an element exists with coordinates matching the supplied values.
     */
    public boolean containsPoint(float x, float y, float z,float w) {
        boolean exists = false;
        for (float[] f: this.set ) {
                if (f[0] == x && f[1] == y && f[2] == z && f[3] == w)
                    exists = true;
            }
        return exists;
    }
    
    /**
     * Removes the object referenced by the parameter.
     * @param point a reference to the float array to be removed.
     * @return true if the object was removed.
     */
    public synchronized boolean remove(float[] point) {
        stats = null;
        return set.remove(point);
    }
    
    /**
     * Removes the object referenced by the parameter.
     * @return true if the point was in this PointSet.
     * @param point a reference to the float array to be removed.
     * @throws sddm.SDDMException if the supplied point is not 3 or 4 elements long.
     */
    public synchronized boolean removePoint(float[] point) throws SDDMException{
        boolean exists = false;
        Iterator<float[]> iter = this.set.iterator();
        
        if (point.length == 3) {
            while(iter.hasNext()) {
                float[] f = iter.next();
                if (f[0] == point[0] &&f[1] == point[1]&&f[2] == point[2]) {
                    iter.remove();
                    stats = null;
                    exists = true;
                }
            }
        }
        else if (point.length == 4) {
            while(iter.hasNext()) {
                float[] f = iter.next();
                if (f[0] == point[0] &&f[1] == point[1]&&f[2] == point[2]&&f[3] == point[3]) {
                    iter.remove();
                    stats = null;
                    exists = true;
                }
            }
        }
        else
            throw new SDDMException("Supplied point to test PointSet for doesn't have 3 or 4 elements.");
        
            return exists;
    }
    
    /**
     * removes the point with the specified coordinates from this PointSet.
     * @param x the x-coordinate to match.
     * @param y the y-coordinate to match.
     * @param z the z-coordinate to match.
     * @return true if the point was in this PointSet.
     */
    public synchronized boolean removePoint(float x, float y, float z) {
        boolean exists = false;
        Iterator<float[]> iter = this.set.iterator();
        while (iter.hasNext()) {
            float[] f = iter.next();
            if (f[0] == x && f[1] == y && f[2] == z) {
                iter.remove();
                exists = true;
                break;
            }
        }
        
        stats = null;
        return exists;
    }
    
    /**
     * removes the point with the specified coordinates from this PointSet.
     * @param x the x-coordinate to match.
     * @param y the y-coordinate to match.
     * @param z the z-coordinate to match.
     * @param w the w-coordinate to match.
     * @return true if the point was in this PointSet.
     */
    public synchronized boolean removePoint(float x, float y, float z,float w) {
        boolean exists = false;
        Iterator<float[]> iter = this.set.iterator();
        while (iter.hasNext()) {
            float[] f = iter.next();
            if (f[0] == x && f[1] == y && f[2] == z && f[3] == w) {
                iter.remove();
                exists = true;
                break;
            }
        }
        
        stats = null;
        return exists;
    }
        
    /**
     * Retrieves the number of elements in this PointSet
     * @return the number of points in this PointSet.
     */
    public int size() {
        return set.size();
    }
    
    /**
     * 
     * Determines if this PointSet is empty.
     * @return true if there are no elements in this PointSet
     */
    public boolean isEmpty() {
        return set.isEmpty();
    }
    
    /**
     * Removes all elements from this PointSet.
     */
    public synchronized void clear() {
        set.clear();
        stats = null;
    }
        
    /**
     * 
     * Generates the iterator for this PointSet.
     * @return an Iterator that references 4-element point arrays.
     */
    public Iterator<float[]> iterator() {
        return set.iterator();
    }
    
    /**
     * 
     * Generates an iterator that returns the x-coordinates of the points in this PointSet.
     * @return an iterator that visits each of the x-coordinates.
     */
    public Iterator<Float> xIterator() {
        return new ElemIter(0);
    }
    
    /**
     * 
     * Generates an iterator that returns the y-coordinates of the points in this PointSet.
     * @return an iterator that visits each of the y-coordinates.
     */
    public Iterator<Float> yIterator() {
        return new ElemIter(1);
    }
    
    /**
     * 
     * Generates an iterator that returns the z-coordinates of the points in this PointSet.
     * @return an iterator that visits each of the z-coordinates.
     */
    public Iterator<Float> zIterator() {
        return new ElemIter(2);
    }
    
    /**
     * 
     * Generates an iterator that returns the w-coordinates of the points in this PointSet.
     * @return an iterator that visits each of the w-coordinates.
     */
    public Iterator<Float> wIterator() {
        return new ElemIter(3);
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     * @param rand the random number generator to use 
     */
    public void randomPerturbation(float scale, boolean doW, Random rand) {
        for (float[] f : this.set) {
            f[0]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            f[1]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            f[2]+=scale*(rand.nextFloat()*2.0f - 1.0f);
            if (doW)
                f[3]+=scale*(rand.nextFloat()*2.0f - 1.0f);
        }
        stats = null;
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     * @param seed the seed to use in generating the default Java RNG
     */
    public void randomPerturbation(float scale, boolean doW, long seed) {
        this.randomPerturbation(scale, doW, new java.util.Random(seed));
    }
    
    /**
     * Randomly adjust each points' coordinates by (at most) the specified amount
     * @param scale the largest allowed perturbation
     * @param doW if flase, does not perturb the w-coordinate
     */
    public void randomPerturbation(float scale, boolean doW) {
        this.randomPerturbation(scale, doW, new java.util.Random());
    }
    
    /**
     * 
     * Retrieves the maximum x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest x value in this PointList.
     */
    public float getXMin() {
        if (stats == null)
            makeStats();
        return stats[0];
    }
    
    /**
     * 
     * Retrieves the minimum x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest x value in this PointList.
     */
    public float getXMax() {
        if (stats == null)
            makeStats();
        return stats[1];
    }
    
    /**
     * 
     * Retrieves the average x value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the x values in this PointList.
     */
    public float getXAvg() {
        if (stats == null)
            makeStats();
        return stats[2];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest y value in this PointList.
     */
    public float getYMin() {
        if (stats == null)
            makeStats();
        return stats[3];
    }
    
    /**
     * 
     * 
     * Retrieves the maximum y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest y value in this PointList.
     */
    public float getYMax() {
        if (stats == null)
            makeStats();
        return stats[4];
    }
    
    /**
     * 
     * 
     * Retrieves the average y value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the y values in this PointList.
     */
    public float getYAvg() {
        if (stats == null)
            makeStats();
        return stats[5];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest z value in this PointList.
     */
    public float getZMin() {
        if (stats == null)
            makeStats();
        return stats[6];
    }
    
    /**
     * 
     * Retrieves the maximum z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest z value in this PointList.
     */
    public float getZMax() {
        if (stats == null)
            makeStats();
        return stats[7];
    }
    
    /**
     * 
     * 
     * Retrieves the average z value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the z values in this PointList.
     */
    public float getZAvg() {
        if (stats == null)
            makeStats();
        return stats[8];
    }
    
    /**
     * 
     * 
     * Retrieves the minimum w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the smallest w value in this PointList.
     */
    public float getWMin() {
        if (stats == null)
            makeStats();
        return stats[9];
    }
    
    /**
     * 
     * Retrieves the maximum w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return the largest w value in this PointList.
     */
    public float getWMax() {
        if (stats == null)
            makeStats();
        return stats[10];
    }
    
    /**
     * 
     * Retrieves the average w value in this PointList.  Will Take longer on the first call after a change has been made.
     * @return The average of the w values in this PointList.
     */
    public float getWAvg() {
        if (stats == null)
            makeStats();
        return stats[11];
    }
    
    /**
     * 
     * Retreives the largest of any of the coordinates in this PointList. (i.e. max of getXMax,getYMax,getZMax,and getWMax)
     * @return the largest value in this PointList.
     */
    public float getMaxMax() {
        if (stats == null)
            makeStats();
        return Math.max(Math.max(stats[10],stats[7]),Math.max(stats[4],stats[1]));
    }
    
    /**
     * 
     * Retreives the smallest of any of the coordinates in this PointList. (i.e. min of getXMin,getYMin,getZMin,and getWMin)
     * @return the smallest value in this PointList.
     */
    public float getMinMin() {
        if (stats == null)
            makeStats();
        return Math.min(Math.min(stats[9],stats[6]),Math.min(stats[3],stats[0]));
    }
    
    /**
     * Determines if these point sets are element-by-element equal.
     * @param obj a PointSet to compare to this (if not, guarenteed to return false)
     * @return true if the elements are all shared and the PointSet 
     */
    public boolean equals(Object obj) {
        boolean b = false;
        if (obj instanceof PointSet) {
            PointSet pso = (PointSet)obj;
            b = this.set.equals(pso.set);
        }
        return b;
    }

    /**
     * Generates a hash code using the backing objects' technique.
     * @return a value fulfilling the hashCode contract.
     */
    public int hashCode() {
        return set.hashCode();
    }
    
    /**
     * Copies this point set element by element.
     * @return a deep copy of this PointSet
     */
    public PointSet clone() {
        PointSet ret = new PointSet();
        for(float[] f : this.set)
            ret.addPoint(f);
        
        if (this.stats != null)
            ret.stats = this.stats.clone();
        else
            ret.stats = null;
        
        return ret; 
    }
    
    /**
     * Generates a list of the points in this PointSet
     * @return a string with all the points on their own line
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("Point Set with Elements:\n");
        
        for ( float[] f : set) {
            sb.append(f[0]);
            sb.append(", ");
            sb.append(f[1]);
            sb.append(", ");
            sb.append(f[2]);
            sb.append(", ");
            sb.append(f[3]);
            sb.append('\n');
        }
        
        return sb.toString();
    }
    
    
    ///-----------------Below are privates-------------------
    private void makeStats() {
        //stats = {xmin,xmax,xavg,ymin,ymax,yavg,zmin,zmax,zavg,wmin,wmax,wavg}
        stats = new float[12];
        
        stats[0] = stats[3] = stats[6] = stats[9] = Float.MAX_VALUE;
        stats[1] = stats[4] = stats[7] = stats[10] = Float.MIN_VALUE;
        stats[2] = stats[5] = stats[8] = stats[11] = 0.0f;
        for (float[] point : this.set) {
            //x-coord
            if (point[0] < stats[0])
                stats[0] = point[0];
            if (point[0] > stats[1])
                stats[1] = point[0];
            stats[2]+=point[0];
            //y-coord
            if (point[1] < stats[3])
                stats[3] = point[1];
            if (point[1] > stats[4])
                stats[4] = point[1];
            stats[5]+=point[1];
            //z-coord
            if (point[2] < stats[6])
                stats[6] = point[2];
            if (point[2] > stats[7])
                stats[7] = point[2];
            stats[8]+=point[2];
            //w-coord
            if (point[3] < stats[9])
                stats[9] = point[3];
            if (point[3] > stats[10])
                stats[10] = point[3];
            stats[11]+=point[3];
        }
        
        //normalize average
        stats[2]/=set.size();
        stats[5]/=set.size();
        stats[8]/=set.size();
        stats[11]/=set.size();
    }

    private class ElemIter implements Iterator<Float>{
        Iterator<float[]> setIter;
        int i;
        
        ElemIter(int index) {
            setIter = set.iterator();
            this.i = index;
        }
        
        public void remove() {
            setIter.remove();
        }
        
        public Float next() {
            return setIter.next()[i];
        }
        
        public boolean hasNext() {
            return setIter.hasNext();
        }
    }
    
}
