/*
 * HeightFieldModule.java
 * Sample command line: -aa 4 -mod  hf -f "C:\Documents and Settings\Erik\mrh.png" i
 */

package sddm.modules;

import java.awt.Color;
import java.awt.event.MouseEvent;
import javax.media.opengl.*;
import javax.swing.*;
import javax.vecmath.*;
import net.java.joglutils.lighting.*;
import sddm.*;
import sddm.SDDModule;
import sddm.data.HeightField;

/**
 *
 * @author Erik J. Tollerud
 */
public class HeightFieldModule extends SDDModule{
    private HeightField hf;
    private int gran;
    private boolean drawPolygons, drawOutline, stipple, lightingOn, drawNormals;
    private float shininess;
    private float scaleFactor;
    private Color outlineColor;
    private Color globalAmbient;
    private Light light;
    private Coloring colorer;
    private SelectListener selectListener;
    private float selectionResult;
    
    private float offset = -4.0f;
    
    /** Creates a new instance of HeightFieldModule */
    public HeightFieldModule() {
        this(new HeightField());
    }
    public HeightFieldModule(StereoCamera cam) {
        this(new HeightField(),cam);
    }
    public HeightFieldModule(HeightField hf) {
        this(hf, new StereoCamera());
    }
        
    public HeightFieldModule(HeightField hf, StereoCamera cam) {
        super(cam);
        this.hf = hf;
        gran = 1;
        this.cam.setCameraLocation(hf.getXSize(),hf.getYSize(),hf.getMaxValue());
        this.cam.setTargetPoint(0,0,0);
        drawPolygons = true;
        drawOutline = false;
        stipple = false;
        lightingOn = true;
        drawNormals = false;
        shininess = 100.0f;
        scaleFactor = 1.0f;
        light = new Light();light.setLightPosition(1,1,1);
        light.setPhongShaded(true);
        outlineColor = new java.awt.Color(0,0,0);
        globalAmbient = new java.awt.Color(0.2f,0.2f,0.2f,1.0f);
        colorer = new Coloring();
        colorer.setTop(hf.getMaxValue());
        colorer.setBottom(hf.getMinValue());
        selectListener = null;
    }
    
    /**
     * Applies selection mode based on the size of the supplied
     * @param drawable the component to base the selection mode on, or null to disable selection
     */
    public void setSelecting(GLAutoDrawable drawable) throws SDDMException {
        if (drawable != null) {
            GLDrawableFactory factory = GLDrawableFactory.getFactory();
            if(!factory.canCreateGLPbuffer())
                throw new SDDMException("Graphics device does not support Pbuffers - can't enter selection mode");
            
            selectListener = new SelectListener(this,drawable.getWidth(),drawable.getHeight()) {
                Coloring oldColor;
                boolean oldLight;
                
                public void storeOld(SDDModule module) {
                    HeightFieldModule hfm = (HeightFieldModule)module;
                    oldColor = hfm.colorer;
                    oldLight = hfm.lightingOn;
                }
                
                public void setupModForSelect(SDDModule module, Coloring indexedColor) {
                    HeightFieldModule hfm = (HeightFieldModule)module;
                    hfm.colorer = indexedColor;
                    hfm.lightingOn = false;
                }
                
                public void reapplyOld(SDDModule module) {
                    HeightFieldModule hfm = (HeightFieldModule)module;
                    hfm.colorer = oldColor;
                    hfm.lightingOn = oldLight;
                    
                }
                
            };
            
        } else 
            selectListener = null;
        
    }
    
    public boolean isSelecting() {
        return (selectListener != null);
    }
    
    public void resize(GL gl, int x, int y, int width, int height) {
        if(isSelecting())
            selectListener.setSize(width,height);
    }
    
    public void recompileDisplayList() {
        super.recompileDisplayList();
        if(isSelecting())
            selectListener.recompile();
    }
    
    public float getSelectionResult() {
        if(!isSelecting())
            throw new SDDMException("Selection mode not active, but attempt was made to get selection result");
        
        return this.selectionResult;
    }
    
    public void setStipple(boolean stipple) {
        this.stipple = stipple;
    }
    
    public boolean isStipple() {
        return stipple;
    }
    
    public void init(GL gl) {
        gl.glPolygonMode(gl.GL_FRONT_AND_BACK,gl.GL_FILL);
        gl.glCullFace(gl.GL_BACK);
        gl.glEnable(gl.GL_CULL_FACE);
        gl.glEnable(gl.GL_DEPTH_TEST);
        light.setLightNumber(0);
        if (lightingOn) {
            try{
                light.enable(gl, 0);
            } catch(LightingException e) {
                JOptionPane.showMessageDialog(null,"Problem setting Phong shader: "+e.getMessage(),"Phong Shading Failed",JOptionPane.WARNING_MESSAGE);
            }
        }
        float[] white = {1.0f,1.0f,1.0f,1.0f};
        gl.glMaterialfv(gl.GL_FRONT_AND_BACK,gl.GL_SPECULAR,white,0);
    }
    
    public void preDraw(GL gl) {
        if(stipple) {
            //gl.glEnable(gl.GL_LINE_SMOOTH);
            //gl.glLineWidth(1000.0f);
            gl.glEnable(gl.GL_LINE_STIPPLE);
            gl.glLineStipple(3, (short)0xAAAA);
        } else {
            gl.glDisable(gl.GL_LINE_STIPPLE);
            //gl.glDisable(gl.GL_LINE_SMOOTH);
        }
        
        if (lightingOn&&light.isPhongShaded())
            light.enable(gl);
    }
    
    public void drawGeometry(GL gl) {
        boolean onRight = this.cam.isRight();
        int xSize = hf.getXSize();
        int ySize = hf.getYSize();
        float sclf = scaleFactor * hf.getZScaleFactor();
        //reset scale for calculating normal only if it's changed
        if (hf.getNormalScaling() != sclf)
            hf.setNormalScaling(sclf);
        
        //generate array to hold vectors for normal calculation
        Vector3f[] vecs = new Vector3f[9];
        for(int i = 0; i < vecs.length;++i)
            vecs[i] = new Vector3f();
        
        if (drawPolygons) {
            //configure lighting
            if (lightingOn) {
                gl.glEnable(gl.GL_LIGHTING);
                try {
                    light.setPhongColorMaterial(true,true);
                    light.enable(gl);
                } catch (LightingException e) {
                    JOptionPane.showMessageDialog(null,"Problem setting Phong shader: "+e.getMessage(),"Phong Shading Failed",JOptionPane.WARNING_MESSAGE);
                }
                gl.glEnable(gl.GL_NORMALIZE);
                gl.glColorMaterial(gl.GL_FRONT_AND_BACK,gl.GL_AMBIENT_AND_DIFFUSE);
                gl.glMaterialf(gl.GL_FRONT_AND_BACK,gl.GL_SHININESS,this.shininess);
                gl.glEnable(gl.GL_COLOR_MATERIAL);
                light.apply(gl);
                Material.applyGlobalAmbient(gl, this.globalAmbient);
            } else {
                gl.glDisable(gl.GL_LIGHTING);
                light.disable(gl);
                gl.glDisable(gl.GL_NORMALIZE);
                gl.glDisable(gl.GL_COLOR_MATERIAL);
            }
            
            float[] RGBColors= {1,1,1};
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK,gl.GL_FILL);
            
            //Draw HeightField mesh
            for(int i = 0; i < (xSize-gran); i+=gran) {
                Vector3f norm;
                //Draw triangle strips one row (fixed x-coord) at a time
                gl.glBegin(gl.GL_TRIANGLE_STRIP);
                for(int j = 0; j < ySize; j+=gran) {
                    float currVal;
                    //Set upper vertex
                    currVal = hf.getData(i,j);
                    //set color
                    gl.glColor3fv(colorer.getColorByValue(currVal),0);
                    //gl.glMaterialfv(gl.GL_FRONT_AND_BACK,gl.GL_AMBIENT_AND_DIFFUSE,colorer.getColorByValue(currVal),0);
                    norm = hf.getNormal(i,j);
                    gl.glNormal3f(norm.x,norm.y,norm.z);
                    gl.glVertex3f(i,j,sclf*currVal);
                    
                    //Set lower vertex
                    currVal = hf.getData(i+gran,j);
                    //set color
                    gl.glColor3fv(colorer.getColorByValue(currVal),0);
                    //gl.glMaterialfv(gl.GL_FRONT_AND_BACK,gl.GL_AMBIENT_AND_DIFFUSE,colorer.getColorByValue(currVal),0);
                    //set normal
                    norm = hf.getNormal(i+gran,j);
                    gl.glNormal3f(norm.x,norm.y,norm.z);
                    
                    gl.glVertex3f(i+gran,j,sclf*currVal);
                }
                gl.glEnd();
            }
            if (lightingOn)
                light.disable(gl);
        }
        if (drawOutline) {
            //don't light the outline
            gl.glDisable(gl.GL_LIGHTING);
            //draw outline
            gl.glEnable(gl.GL_POLYGON_OFFSET_LINE);
            gl.glPolygonOffset(offset,1);
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK,gl.GL_LINE);
            gl.glColor3fv(this.outlineColor.getRGBColorComponents(null),0);
            gl.glLineWidth(1);
            for(int i = 0; i < xSize-gran; i+=gran) {
                gl.glBegin(gl.GL_TRIANGLE_STRIP);
                for(int j = 0; j < ySize; j+=gran) {
                    gl.glVertex3f(i,j,sclf*hf.getData(i,j));
                    gl.glVertex3f(i+gran,j,sclf*hf.getData(i+gran,j));
                }
                gl.glEnd();
            }
            gl.glDisable(gl.GL_POLYGON_OFFSET_LINE);
        }
        if (drawNormals) {
            gl.glDisable(gl.GL_LIGHTING);
            //use outline color
            gl.glColor3fv(this.outlineColor.getRGBColorComponents(null),0);
            gl.glBegin(gl.GL_LINES);
            for(int i = 0; i < xSize; i+=gran) {
                for(int j = 0; j < ySize; j+=gran) {
                    float currVal = hf.getData(i,j);
                    Vector3f norm = hf.getNormal(i,j);
                    gl.glVertex3f(i,j,sclf*currVal);
                    gl.glVertex3f(norm.x+(float)i,norm.y+(float)j,norm.z+sclf*currVal);
                    
                }
            }
            gl.glEnd();
        }
    }
    
    public static String moduleTitle() {
        return "Height Field Module";
    }
    
    public HeightFieldModuleControlPanel newModuleControlPanel() {
        return new HeightFieldModuleControlPanel(this,cam);
    }
    
    public static HeightFieldModuleSetupPanel newStartupPanel() {
        return new HeightFieldModuleSetupPanel();
    }
    
    public void setGranularity(int granularity) {
        if (granularity > 0)
            this.gran = granularity;
    }
    
    public boolean isDrawOutline() {
        return drawOutline;
    }
    
    public void setDrawOutline(boolean drawOutline) {
        this.drawOutline = drawOutline;
        this.recompileDisplayList();
    }
    
    public boolean isDrawPolygons() {
        return drawPolygons;
    }
    
    public void setDrawPolygons(boolean drawPolygons) {
        this.drawPolygons = drawPolygons;
        this.recompileDisplayList();
    }
    public int getGranularity() {
        return this.gran;
    }
    public void attachEventListeners(StereoCameraControlFrame frame) {
        super.attachEventListeners(frame);
        final StereoCameraControlFrame frameF = frame;
        frame.addKeyListenerToJFrames(new java.awt.event.KeyAdapter() {
            public void keyTyped(java.awt.event.KeyEvent e) {
                switch(e.getKeyChar()) {
                    case 'o':
                        drawOutline = !drawOutline;
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                    case 'p':
                        drawPolygons = !drawPolygons;
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                    case 't':
                        stipple = stipple ? false : true;
                        frameF.redisplay();
                        break;
                    case 'O':
                        String res = JOptionPane.showInputDialog("Input new poygon offset dist");
                        offset = Float.parseFloat(res);
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                    case 'n':
                        drawNormals = !drawNormals;
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                    case 'x':
                        if(isDrawAxis() && getNumOfTicks() > 0)
                            setDrawAxis(false,0,0,0);
                        else if (isDrawAxis() && getNumOfTicks() == 0)
                            setDrawAxis(true,hf.getXSize(),hf.getYSize(),hf.getMaxValue()*scaleFactor*hf.getZScaleFactor(),5);
                        else
                            setDrawAxis(true,hf.getXSize(),hf.getYSize(),hf.getMaxValue()*scaleFactor*hf.getZScaleFactor(),0);
                        
                        frameF.redisplay();
                        break;
                    case 'X':
                        //Aligns along axes
                        float axisScale = (hf.getXSize() > hf.getYSize()) ? hf.getXSize() : hf.getYSize();
                        if(hf.getMaxValue()*scaleFactor*hf.getZScaleFactor() > axisScale)
                            axisScale = hf.getMaxValue()*scaleFactor*hf.getZScaleFactor();
                        Vector3f loc = cam.getCameraLocation();
                        if (loc.x == axisScale && loc.y == 0 && loc.z == 0) {
                            cam.setCameraLocation(0,axisScale,0);
                            cam.setUpVector(0,0,1);
                        } else if (loc.x == 0 && loc.y == axisScale && loc.z == 0) {
                            cam.setCameraLocation(0,0,axisScale);
                            cam.setUpVector(0,1,0);
                        } else {
                            cam.setCameraLocation(axisScale,0,0);
                            cam.setUpVector(0,1,0);
                        }
                        cam.setTargetPoint(0,0,0);
                        frameF.redisplay();
                        break;
                    case 'l':
                        lightingOn = !lightingOn;
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                    case 'c':
                        float weight = Float.parseFloat(JOptionPane.showInputDialog("Convolution Weight?"));
                        boolean b = (JOptionPane.showConfirmDialog(null,"Convolve with Corners?","",JOptionPane.YES_NO_OPTION)==JOptionPane.YES_OPTION);
                        hf.smoothConvolution(weight,b);
                        recompileDisplayList();
                        frameF.redisplay();
                        break;
                        
                }
                frameF.getSecondPanel().moduleSettingsToPanel(HeightFieldModule.this);
                frameF.repaint();
            }
        });
        
        frame.addMouseListenerToJFrames(new java.awt.event.MouseAdapter() {
            public void mouseClicked(MouseEvent evt) {
                if(isSelecting()) {
                    try {
                        selectionResult = selectListener.select(evt.getX(),evt.getY());
                        frameF.redisplay();
                    } catch (InterruptedException e) {
                        System.err.println(e.getMessage());
                        selectionResult = 0;
                    }
                    
                }
            }
        });
    }
    
    public Color getOutlineColor() {
        return outlineColor;
    }
    
    public void setOutlineColor(Color outlineColor) {
        this.outlineColor = outlineColor;
    }
    
    public Light getLight() {
        return this.light;
    }
    
    HeightField getField() {
        return this.hf;
    }
    
    public boolean isLightingOn() {
        return lightingOn;
    }
    
    public void setGlobalAmbient(Color globalAmbient) {
        this.recompileDisplayList();
        this.globalAmbient = globalAmbient;
    }
    
    public Color getGlobalAmbient() {
        return globalAmbient;
    }
    
    public void setLightingOn(boolean lightingOn) {
        if(lightingOn != this.lightingOn)
            this.recompileDisplayList();
        this.lightingOn = lightingOn;
    }
    
    public void setDrawNormals(boolean drawNormals) {
        this.recompileDisplayList();
        this.drawNormals = drawNormals;
    }
    
    public boolean isDrawNormals() {
        return drawNormals;
    }
    
    public void setShininess(float shininess) throws SDDMException {
        if (shininess >= 0 && shininess <=128) {
            this.shininess = shininess;
            this.recompileDisplayList();
        } else
            throw new SDDMException("tried to set shininess outside [0,128]");
    }
    
    public float getShininess() {
        return shininess;
    }
    
    public static SDDModule makeModuleFromArgs(String modName, java.util.HashMap<String,Integer> switchMap, String[] args, StereoCamera camera) throws  ParsingException{
        //takes modName hf or heightfield
        //accepts these switches:
        //-file|f (filename|"prompt") (image|i)|(ascii1d|1d|1)|(ascii2d|2d|2))
        // | -fprompt|fileprompt
        //
        //[-lighting] (true|t)|(false|f)
        //[-outline] (true|t)|(false|f)
        //[-mesh|polygons|polygonmesh] (true|t)|(false|f)
        //[-normals] (true|t)|(false|f)
        //[-stipple] (true|t)|(false|f)
        //[-granularity|gran] #
        //[-bgcolor] r g b
        //[-outlinecolor] r g b
        //[-specular] r g b
        //[-diffuse] r g b
        //[-ambient] r g b
        //[-globalambient] r g b
        //[-shininess] #.#
        //[-phongShading|phong] (true|t)|(false|f)
        //[-zscale|verticalscale] #.#
        //[Coloring switches]
        
        
        if (modName.matches("hf|heightfield")) {
            HeightFieldModule m = null;
            Integer fileIndx = switchMap.remove("-f");
            if (fileIndx == null)
                fileIndx = switchMap.remove("-file");
            if (fileIndx != null) {
                String fnStr = args[fileIndx+1];
                java.io.File file;
                if (fnStr.equalsIgnoreCase("prompt")) {
                    JFileChooser jfc = new JFileChooser();
                    jfc.setFileSelectionMode(jfc.FILES_ONLY);
                    jfc.setMultiSelectionEnabled(false);
                    
                    jfc.setDialogType(jfc.OPEN_DIALOG);
                    if(jfc.APPROVE_OPTION == jfc.showDialog(null,null))
                        file = jfc.getSelectedFile();
                    else
                        throw new ParsingException("Cancelled file dialog while generating height field");
                    
                } else
                    file = new java.io.File(fnStr);
                
                String format = args[fileIndx+2];
                HeightField.hfFileType type = HeightField.hfFileType.ASCII_1D;
                if (format.matches("image|i"))
                    type = HeightField.hfFileType.IMAGE;
                else if (format.matches("ascii1d|1d|1"))
                    type = HeightField.hfFileType.ASCII_1D;
                else if (format.matches("ascii2d|2d|2"))
                    type = HeightField.hfFileType.ASCII_2D;
                else
                    throw new ParsingException("Unrecognized height field file format:"+type);
                
                try {
                    m = new HeightFieldModule(new HeightField(file,type),camera);
                } catch (java.io.IOException e) {
                    throw new ParsingException("IOException while loading height field:\n"+e.getMessage(),e);
                } catch (FileFormatException e) {
                    throw new ParsingException("FileFormatException while loading height field:\n"+e.getMessage(),e);
                }
            }
            Integer fpIndx = switchMap.remove("-fprompt");
            if (fpIndx == null)
                fpIndx = switchMap.remove("-fileprompt");
            if (fpIndx != null) {
                JFileChooser jfc = new JFileChooser();
                jfc.setFileSelectionMode(jfc.FILES_ONLY);
                jfc.setMultiSelectionEnabled(false);
                
                jfc.setDialogType(jfc.OPEN_DIALOG);
                
                if(jfc.APPROVE_OPTION == jfc.showDialog(null,null)) {
                    java.io.File file = jfc.getSelectedFile();
                    HeightField.hfFileType format = HeightField.hfFileType.ASCII_2D;
                    if(JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(null,"Image file format?","File Format Type",JOptionPane.YES_NO_OPTION))
                        format = HeightField.hfFileType.IMAGE;
                    if(format == HeightField.hfFileType.ASCII_2D && JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(null,"1D format?","File Format Type",JOptionPane.YES_NO_OPTION))
                        format = HeightField.hfFileType.ASCII_1D;
                    try {
                        m = new HeightFieldModule(new HeightField(file,format),camera);
                    } catch (java.io.IOException e) {
                        throw new ParsingException("IOException thrown while making HeightField:"+e.getMessage(),e);
                    } catch (FileFormatException e) {
                        throw new ParsingException("ParsingException thrown while making HeightField:"+e.getMessage(),e);
                    }
                } else
                    throw new ParsingException("Cancelled file dialog while generating height field");
            }
            
            
            if (m== null)
                throw new ParsingException("No field provided for height field module");
            
            Integer lightingIndx = switchMap.remove("-lighting");
            if (lightingIndx != null) {
                String lightStr = args[lightingIndx+1];
                if (lightStr.matches("t|true"))
                    m.setLightingOn(true);
                if (lightStr.matches("f|false"))
                    m.setLightingOn(false);
            }
            
            Integer olIndx = switchMap.remove("-outline");
            if (olIndx != null) {
                String olStr = args[olIndx+1];
                if (olStr.matches("t|true"))
                    m.setDrawOutline(true);
                if (olStr.matches("f|false"))
                    m.setDrawOutline(false);
            }
            
            Integer meshIndx = switchMap.remove("-mesh");
            if (meshIndx == null)
                meshIndx = switchMap.remove("-polygonmesh");
            if (meshIndx == null)
                meshIndx = switchMap.remove("-polygons");
            if (meshIndx != null) {
                String meshStr = args[meshIndx+1];
                if (meshStr.matches("t|true"))
                    m.setDrawPolygons(true);
                if (meshStr.matches("f|false"))
                    m.setDrawPolygons(false);
            }
            
            Integer normIndx = switchMap.remove("-normals");
            if (normIndx != null) {
                String normStr = args[normIndx+1];
                if (normStr.matches("t|true"))
                    m.setDrawNormals(true);
                if (normStr.matches("f|false"))
                    m.setDrawNormals(false);
            }
            
            Integer stipIndx = switchMap.remove("-stipple");
            if (stipIndx != null) {
                String stipStr = args[stipIndx+1];
                if (stipStr.matches("t|true"))
                    m.setDrawNormals(true);
                if (stipStr.matches("f|false"))
                    m.setDrawNormals(false);
            }
            
            Integer granIndx = switchMap.remove("-gran");
            if (granIndx == null)
                granIndx = switchMap.remove("-granularity");
            if (granIndx != null) {
                int g = Integer.parseInt(args[granIndx+1]);
                m.setGranularity(g);
            }
            
            Integer bgIndx = switchMap.remove("-bgcolor");
            if (bgIndx != null) {
                float r = Float.parseFloat(args[bgIndx+1]);
                float g = Float.parseFloat(args[bgIndx+2]);
                float b = Float.parseFloat(args[bgIndx+3]);
                m.setBackgroundColor(new Color(r,g,b));
            }
            
            Integer olcIndx = switchMap.remove("-outlinecolor");
            if (olcIndx != null) {
                float r = Float.parseFloat(args[olcIndx+1]);
                float g = Float.parseFloat(args[olcIndx+2]);
                float b = Float.parseFloat(args[olcIndx+3]);
                m.setOutlineColor(new Color(r,g,b));
            }
            
            Integer specIndx = switchMap.remove("-specular");
            if (specIndx != null) {
                float r = Float.parseFloat(args[specIndx+1]);
                float g = Float.parseFloat(args[specIndx+2]);
                float b = Float.parseFloat(args[specIndx+3]);
                m.light.setSpecular(new Color(r,g,b));
            }
            
            Integer diffIndx = switchMap.remove("-diffuse");
            if (diffIndx != null) {
                float r = Float.parseFloat(args[diffIndx+1]);
                float g = Float.parseFloat(args[diffIndx+2]);
                float b = Float.parseFloat(args[diffIndx+3]);
                m.light.setDiffuse(new Color(r,g,b));
            }
            
            Integer ambIndx = switchMap.remove("-ambient");
            if (ambIndx != null) {
                float r = Float.parseFloat(args[ambIndx+1]);
                float g = Float.parseFloat(args[ambIndx+2]);
                float b = Float.parseFloat(args[ambIndx+3]);
                m.light.setAmbient(new Color(r,g,b));
            }
            
            Integer glAmbIndx = switchMap.remove("-globalambient");
            if (glAmbIndx != null) {
                float r = Float.parseFloat(args[glAmbIndx+1]);
                float g = Float.parseFloat(args[glAmbIndx+2]);
                float b = Float.parseFloat(args[glAmbIndx+3]);
                m.setGlobalAmbient(new Color(r,g,b));
            }
            
            Integer shinIndx = switchMap.remove("-shininess");
            if (shinIndx != null) {
                float shininess = Float.parseFloat(args[shinIndx+1]);
                m.setShininess(shininess);
            }
            
            Integer phongIndx = switchMap.remove("-phongShading");
            if (phongIndx == null)
                phongIndx = switchMap.remove("-phong");
            if (phongIndx != null) {
                String phongStr = args[phongIndx+1];
                
                if (phongStr.matches("t|true")) {
                    try{
                        m.getLight().setPhongShaded(true);
                    } catch (LightingException e) {
                        JOptionPane.showMessageDialog(null,"Problem setting Phong shader: "+e.getMessage(),"Phong Shading Failed",JOptionPane.WARNING_MESSAGE);
                    }
                } else if(phongStr.matches("f|false"))
                    m.getLight().setPhongShaded(false);
                else
                    throw new ParsingException("Invalid switch supplied for phong shading: "+phongStr);
            }
            
            Integer zScaleIndx = switchMap.remove("-zscale");
            if (zScaleIndx == null)
                zScaleIndx = switchMap.remove("-verticalscale");
            if (zScaleIndx != null) {
                float scale = Float.parseFloat(args[zScaleIndx+1]);
                m.setZScale(scale);
            }
            
            Coloring c = Coloring.makeColoringFromArgs(switchMap, args);
            if( c != null )
                m.setColorer(c);
            
            
            //TODO: add new switches' effects here
            
            return m;
        }
        return null;
    }
    
    public float getMinFeatureSizeHint() {
        
        if (hf == null)
            return super.getMinFeatureSizeHint();
        return (hf.getXSize() + hf.getYSize())/1000.0f;
    }
    
    public void setZScale(float zScale) {
        this.recompileDisplayList();
        this.scaleFactor = zScale/hf.getZScaleFactor();
        hf.setNormalScaling(this.scaleFactor);
    }
    
    public float getZScale() {
        return hf.getZScaleFactor()*scaleFactor;
    }
    
    public Coloring getColorer() {
        return this.colorer;
    }
    
    public void setColorer(Coloring colorer) {
        this.colorer = colorer;
        this.colorer.setTop(hf.getMaxValue());
        this.colorer.setBottom(hf.getMinValue());
    }
    
    /*private class SelectListener implements GLEventListener {
        boolean recompile;
        boolean selectionNeeded;
        Integer selectDisplayList;
        float result;
        int selectX;
        int selectY;
        Coloring indexColorer;
        private GLPbuffer selectPbuffer;
        
        
        SelectListener(GLPbuffer buffer) {
            selectPbuffer = buffer;
            selectDisplayList = null;
            selectionNeeded = false;
            recompile = true;
            result = 0;
            selectX=1;
            selectY=1;
            indexColorer = new Coloring();
            indexColorer.setColorSchemeToIndexValue();
            selectPbuffer.addGLEventListener(this);
        }
        
        
        public void init(GLAutoDrawable drawable) {
            GL gl = drawable.getGL();
            selectDisplayList = gl.glGenLists(1);
            //TODO:test if this is enough
            HeightFieldModule.this.init(gl);
        }
        
        public void display(GLAutoDrawable drawable) {
            GL gl = drawable.getGL();
            if (recompile) {
                Coloring oldColor = HeightFieldModule.this.colorer;
                boolean oldLight = HeightFieldModule.this.lightingOn;
                
                try {
                    //Set color and scale for selection
                    HeightFieldModule.this.colorer = indexColorer;
                    HeightFieldModule.this.lightingOn = false;
                    
                    gl.glNewList(selectDisplayList,GL.GL_COMPILE);
                    HeightFieldModule.this.drawGeometry(gl);
                    gl.glEndList();
                } catch(Exception e) {
                    throw new SDDMException("Problem during selection display list generation:"+e.getMessage(),e);
                } finally {
                    //return color and scale to original values
                    HeightFieldModule.this.colorer = oldColor;
                    HeightFieldModule.this.lightingOn = oldLight;
                }
            }
            if (selectionNeeded) {
                
                gl.glClearColor(0.0f,0.0f,0.0f,0.0f);
                gl.glClear(GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT);
                //TODO:test drawing
                
                cam.updateCamera(gl);
                //update again to preserve stereo order for other modules TODO:major fix here to account for selecting both eyes
                if(cam.isStereo())
                    cam.updateCamera(gl);
                gl.glCallList(selectDisplayList);
                
                //read out the selected pixel, get the value, and apply it to the array
                java.nio.FloatBuffer buff = com.sun.opengl.util.BufferUtil.newFloatBuffer(4);
                gl.glReadPixels(selectX,selectY,1,1,GL.GL_RGBA,GL.GL_FLOAT,buff);
                result = indexColorer.getIndexedFloat(buff.get(),buff.get(),buff.get(),buff.get());
                //selection completed - wait for another request before doing the work again
                selectionNeeded = false;
            }
        }
        
        /**
         * Call this, then repaint the component in question to get selected value stored in valueHolder
         * 
         * @param x the x-coordinate to select in window coordinates
         * @param y the y-coordinate to select in window coordinates
         * @throws java.lang.InterruptedException if the thread is interrupted while waiting for selection to be completed
         * @return the selected value at that location
         */
        /*public float select(int x, int y) throws InterruptedException{
            //TODO: figure out why this isn't working - maybe something's funky elsewhere?
            selectionNeeded = true;
            selectX = x;
            selectY = selectPbuffer.getHeight()-y;
            selectPbuffer.repaint();
            while (selectionNeeded)
                wait(100);
            return result;
        }
        
        public void setSize(int width, int height) {
            selectPbuffer.setSize(width,height);
        }

        public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
        }
        public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {
        }
        
        
    }*/
    
}
